/**
 * Copyright © 2013 by Data Virtuality GmbH
 */
package com.datavirtuality.dv.core.interfaces;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import java.util.Set;
import javax.sql.DataSource;




/**
 * Interface for a generic DWH platform
 * 
 * @author Salvatore Raunich
 * @date 2014.09.03
 */
public interface DwhPlatform {
	public Object createPlatformInstance(DataSource ds);

	public Object getPlatform(String schemaNameInDV);

	public Object getPlatform(boolean delimitedIdentifierModeOn, String schemaNameInDV);

	public DataSource getDataSource(String name);

	public Object getExecutionFactory();

	public void setExecutionFactory(Object ef);

	public boolean supportsIndexes();

	public boolean supportsSortKeys();
	
	public boolean supportsDistKeys();

	public String getSQLForGatheringNativeStatistics(String fullyQualifiedTableNameInSource);

	/**
	 * Updates the native type mapping
	 */
	public void fixNativeTypes(Object platform);

	public <T extends Exception> String createTable(String targetTableName, Object tableMetadata, String schemaName, Map<String, String> optionalCreationParams, Class<? extends T> clazz, Object connection) throws T;

	public <T extends Exception> String createTable(String targetTableName, Object tableMetadata, String schemaName, Map<String, String> optionalCreationParams, Class<? extends T> clazz, Object connection, boolean generateScriptOnly) throws T;

	public <T extends Exception> String createTable(String targetTableName, List<? extends Object> objects, String schemaName,
													Class<? extends T> clazz, Object connection, Object createTableCallback, boolean generateScriptOnly) throws T;

	public <T extends Exception> String createTable(String targetTableName, List<? extends Object> objects, String schemaName, Map<String, String> optionalCreationParams,
													Class<? extends T> clazz, Object connection, Object createTableCallback, boolean generateScriptOnly) throws T;
	
	public void dropTable(String tableName, String targetSchemaNameInTeiid, String schemaNameOriginal, boolean printException, int dropTimeout) throws Exception;

	public void dropTable(String tableName, String targetSchemaNameInTeiid, String schemaNameOriginal) throws Exception;

	public List<String> getAlterTableSql(Object targetTable, Object tableMetadata, final String schemaName) throws Exception;

	public List<String> getAlterColumnsSql(Object targetTable, Object tableMetadata, String template) throws Exception;

	/**
	 * Generates a name unique name for a column/index.
	 * 
	 * @param sourceColumnShortName
	 *            the short name of the source column
	 * @param names
	 *            a map of names
	 * @return the quoted generated column name
	 */
	public String generateColumnName(Object platform, String sourceColumnShortName, Map<String, String> names);

	/**
	 * takes a non-quoted column name and returns it quoted with rdbms-specific Quotes
	 * 
	 * @param nonQuotedColumnName
	 *            non-quoted column name like "column1"
	 * @param escapeQuotes
	 *            if quotes should be escaped or not
	 * @param schemaName
	 *            the target schema name
	 * @return quoted column name like "`column1`"
	 */
	public String toQuotedColumnName(String nonQuotedColumnName, boolean escapeQuotes, String schemaName);

	/**
	 * takes a non-quoted column name and returns it quoted with rdbms-specific Quotes
	 * 
	 * @param nonQuotedColumnName
	 *            non-quoted column name like "column1"
	 * @param schemaName
	 *            the target schema name
	 * @return quoted column name like "`column1`"
	 */
	public String toQuotedColumnName(String nonQuotedColumnName, String schemaName);

	/**
	 * takes a quoted identifier and returns it quoted with rdbms-specific Quotes
	 * 
	 * @param quotedColumnName
	 *            quoted identifier
	 * @param schemaName
	 *            the target schema name
	 * @return unquoted identifier
	 */
	public String toUnquotedName(String quotedColumnName, String schemaName);

	public boolean createIndex(Object db, String indexNameSuffix, List<String> columnNames, String expression,
			Object dwhTable, Object table, String targetSchemaNameOriginal, String targetSchemaNameInDV, boolean exceptionOnBadColumn) throws Exception;

	public void dropIndex(Object matTable, Object recIndexObject, String schemaName) throws Exception;

	/**
	 * execute a SQL query
	 * 
	 * @param sql
	 *            the SQL query to execute
	 * @throws Exception
	 *             if an error occurs executing the query
	 */
	public void executeQuery(String sql) throws Exception ;

	/**
	 * get the catalog name
	 * 
	 * @return the catalog name
	 */
	String getCatalog(Object platform);

	String getSchema(Object platform);

	public Object prepareModelWithTable(String schemaOriginal, String schemaDV, String tableName);

	public String determineSchemaOf(String schemaDV, String tableName, List<String> columnNames);

	String caseIndexExtraProcessing(Object expr, String nativeExpr);

	public String escapeStringValue(String value);

	public void setMaxTableNameLangth(int value);

	public int getMaxTableNameLength();

	public void setMaxColumnNameLength(int value);
	
	public int getMaxColumnNameLength();
	
	public void setDefaultSchema(String defaultSchema);

	public String getDefaultSchema();

	public String getDefaultSchemaOrThrow() throws Exception;

	String mapColumn(Object matTable, int sqlType, Object column, Object col);

	public boolean requiresDataSource();

	public void setName(String name);

	public String getName();

	public String getShortTableName(String fullyQualifiedName);
	
	public Collection<String> getSupportedOptions();
	
	public void addTableDefForCreateTableAsSelect(Object selectInto, Object visitor);

	public String generateCreationalParams(Object selectInto);

	String setTableNameCaseType(String tableName);

	public <T extends Exception> TableNameInSource getTableNameInSource(Object platform, String tableName, Class<? extends T> exc) throws T;

	public <T extends Exception> TableNameInSource parseTableNameInSource(Object platform, String tableName, Class<? extends T> exc) throws T;

	public TableNameInSource getTableNameInSource(Object platform, String schemaName, String tableName);

	String getTableNameByDwhPlatform(DwhPlatform dwhPlatform, String schemaName, String tableName);

	public String normalizeTableName(String tableName);
	
	void setUseFullSchemaName(boolean useFullSchemaName);

	void setUseCatalogName(boolean useCatalogName);

	void setSchemaPatterns(Set<String> schemaPatterns);

	void setExcludeSchemas(Set<String> excludeSchemas);
	
	boolean isUseFullSchemaName();

	boolean isUseCatalogName();
	
	void setCreateStringsAsClobs(boolean createStringsAsClobs);
	
	boolean isCreateStringsAsClobs();

	void setStringClobsLengthThreshold(int stringClobsLengthThreshold);

	int getStringClobsLengthThreshold();

	/**
	 * Validates the options specified in Select into command
	 * (So far only for Redshift platform) 
	 * 
	 * @throws Exception
	 */
	public void validateOptions(Object table, Map<String, String> params) throws Exception;

	public boolean validateTypesForInsert(String dataType, String columnType);
	
	public Class<?> fixColumnType(Object column, Object otherCol);
	
	public String fixRuntimeType(Object column);

	public boolean supportsCreateTableAsSelect();

	public int setColumnProps(Object col, Object column, boolean alter);

	void setTableColumnsProps(Object table);

	public void buildNativeType(Object platformInfo, Object col);

	boolean isSchemaAllowed(Object platform, String schemaName);

	boolean isTableExists(String teiidSchema, String defaultSchema, String tableName);

	String getTruncateSql(String teiidSchema, String fullyQualifiedTableNameInSource);

    void recreateMappings(Object matTable, Object matchDescr, Object table);

    class TableNameInSource {
		private String defaultSchema;
		private String schemaByPlatform;
		private char delim;
		private String tableName;
		private String quote;
		private String prefix;
		private boolean useFullSchemaName;
		private boolean useCatalogName;
		private final boolean useSchemaInSource;
		private boolean customDefinedCatalog;

		public TableNameInSource(DwhPlatform dwhPlatform, String defaultSchema, String schemaByPlatform, char delim, String tableName, String quote, String prefix, boolean useFullSchemaName, boolean useCatalogName, boolean useSchemaInSource, boolean customDefinedCatalog) {
			this.defaultSchema = defaultSchema;
			this.schemaByPlatform = (schemaByPlatform == null || "".equals(schemaByPlatform)) ? "" : schemaByPlatform;
			this.delim = delim;
			this.tableName = tableName;
			this.quote = quote;
			this.prefix =  (prefix == null || "".equals(prefix)) ? "" : prefix;
			this.useFullSchemaName = useFullSchemaName;
			this.useCatalogName = useCatalogName;
			this.useSchemaInSource = useSchemaInSource;
			this.customDefinedCatalog = customDefinedCatalog;
		}

		public String getPrefix(boolean quote) {
			return quote ? getPrefix() : prefix;
		}

		public String getSchema(boolean quote) {
			return quote ? getSchema() : defaultSchema;
		}

		public String getSchemaByPlatform(boolean quote) {
			return quote ? getSchemaByPlatform() : schemaByPlatform;
		}

		public String getTableName(boolean quote) {
			return quote ? getTableName() : tableName;
		}

		public String getPrefix() {
			return prefix.length() > 0 ? quote + prefix + quote : prefix;
		}

		public String getSchema() {
			return quote + defaultSchema + quote;
		}

		public String getSchemaByPlatform() {
			return schemaByPlatform.length() > 0 ? quote + schemaByPlatform + quote : schemaByPlatform;
		}

		public String getSchema(boolean quote, boolean defaultSchema) {
			String schemaByPlatform = getSchemaByPlatform(quote);
			String schema = getSchema(quote);
			// case when platform supports schemas + schemaByPlatform and defaultSchema are set -
			// then we should choose defaultSchema because it can be set explicitly outside via tableName
			boolean shouldUseDefaultSchema = schemaByPlatform.length() > 0 && !schema.equals(schemaByPlatform);

			return (defaultSchema || shouldUseDefaultSchema ?
					schema : (schemaByPlatform.length() > 0 ? schemaByPlatform : ""));
		}

		public String getTableName() {
			return quote + tableName + quote;
		}

		public String getFullName() {
			return getSchema() + delim + getTableName();
		}

		public String getFullName(boolean quote) {
			return getFullName(quote, false, true);
		}

		public String getFullName(boolean quote, boolean withPrefix, boolean defaultSchema) {
			String prefix = withPrefix ? getPrefix(quote) : "";
			String schema = useSchemaInSource ? getSchema(quote, defaultSchema) : "";
			return (prefix.length() > 0 ? prefix + delim : "") + (schema.length() > 0 ? schema + delim : "") +
					getTableName(quote);
		}

		public String getTableNameByDwhPlatform() {
			return useFullSchemaName ? getFullName(false, useCatalogName, false) : getTableName(false);
		}

		public String getTableNameByDwhPlatformInSource() {
			return getFullName(true, useCatalogName, false);
		}

		public boolean isCustomDefinedCatalog() {
			return customDefinedCatalog;
		}

		@Override
		public String toString() {
			return "TableNameInSource{" +
					"defaultSchema='" + defaultSchema + '\'' +
					", schemaByPlatform='" + schemaByPlatform + '\'' +
					", delim=" + delim +
					", tableName='" + tableName + '\'' +
					", quote='" + quote + '\'' +
					", prefix='" + prefix + '\'' +
					", useFullSchemaName=" + useFullSchemaName +
					", useCatalogName=" + useCatalogName +
					", useSchemaInSource=" + useSchemaInSource +
					'}';
		}


	}
}
