/**
 * Copyright © 2017 by Data Virtuality GmbH
 */
package com.datavirtuality.dv.core.interfaces.concurrent;

import java.lang.ref.WeakReference;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.Semaphore;

/**
 * A dual semaphore implementation.
 * 
 * Imagine a rail crossing. When a train goes, no other train may go, neither cars.
 * Otherwise, when no train goes, several cars may go, but a train also has to wait for cars crossing.
 * 
 * @author Roman Pedchenko
 * @date 2013.05.17
 */
public class DualSemaphore {
	private final class WeakReferenceComparable extends WeakReference<Thread> implements Comparable<WeakReferenceComparable> {
		private WeakReferenceComparable(Thread referent) {
			super(referent);
		}

		@Override
		public int compareTo(WeakReferenceComparable o) {
			Thread t1 = get();
			Thread t2 = o.get();
			int th1 = t1 != null ? t1.hashCode() : 0;
			int th2 = t2 != null ? t2.hashCode() : 0;
			return th1 - th2;
		}
	}

	private Semaphore trainSemaphore = new Semaphore(1, true);
	private Semaphore carSemaphore = new Semaphore(1, true);
	private Set<WeakReference<Thread>> carSet = new ConcurrentSkipListSet<>();
	private String schemaName;

	public DualSemaphore(String schemaName) {
		this.schemaName = schemaName;
	}

	public boolean acquireTrain() {
		boolean result = trainSemaphore.availablePermits() > 0;
		trainSemaphore.acquireUninterruptibly();
		carSemaphore.acquireUninterruptibly();
		return result;
	}

	public void releaseTrain() {
		carSemaphore.release();
		trainSemaphore.release();
	}

	public void acquireCar(Thread thread) {
		trainSemaphore.acquireUninterruptibly();
		carSemaphore.tryAcquire();
		carSet.add(new WeakReferenceComparable(thread));
		trainSemaphore.release();
	}

	public void releaseCar(Thread thread) {
		boolean found = false;
		for (Iterator<WeakReference<Thread>> i = carSet.iterator(); i.hasNext();) {
			WeakReference<Thread> v = i.next();
			if (v.get() == null) {
				i.remove();
				continue;
			}
			if (!found && v.get() == thread) {
				found = true;
				i.remove();
			}
		}
		if (!found) {
			return;
		}
		if (carSet.size() == 0)
			carSemaphore.release();
	}

	@Override
	public String toString() {
		return super.toString() + "[ Permits: schema: " + schemaName + "; train: " + trainSemaphore.availablePermits() + ", cars: " + carSemaphore.availablePermits() + "; Car count: " + carSet.size() + " ]";
	}
}
